from dataclasses import dataclass, field


@dataclass
class DynamicTopPriceModel:
    three: list = field(default_factory=lambda: [])
    seven: list = field(default_factory=lambda: [])
    thirty: list = field(default_factory=lambda: [])
    month: list = field(default_factory=lambda: [])



import logging
from abc import abstractmethod
from datetime import date, datetime
from statistics import mean
from peaqevcore.common.spotprice.const import AVERAGE_MAX_LEN
from peaqevcore.common.models.observer_types import ObserverTypes
from peaqevcore.common.spotprice.dynamic_top_price import DynamicTopPrice
from peaqevcore.common.spotprice.models.spotprice_dto import ISpotPriceDTO
from peaqevcore.common.spotprice.models.spotprice_model import SpotPriceModel
from peaqevcore.common.models.peaq_system import PeaqSystem

_LOGGER = logging.getLogger(__name__)


class MockObserver:
    async def async_broadcast(self, type: ObserverTypes, data = None) -> None:
        pass


class SpotPriceBase:
    def __init__(self, hub, source: str, system: PeaqSystem, observer = MockObserver(), test:bool = False, is_active: bool = True):
        _LOGGER.debug(f"Initializing Spotprice for {source} from system {system.value}.")
        self.hub = hub
        self.observer = observer
        self.model = SpotPriceModel(source=source)
        self._dynamic_top_price = DynamicTopPrice()
        self._is_initialized: bool = False

        self.converted_average_data: bool = False #remove this five versions from peaqev 3.2.0
        if not test:
            self.state_machine = hub.state_machine
            if is_active:
                self.setup()

    @property
    def tomorrow_valid(self) -> bool:
        return getattr(self.model, "tomorrow_valid", False)

    @property
    def entity(self) -> str:
        return getattr(self.model, "entity", "")

    @property
    def is_initialized(self) -> bool:
        return self._is_initialized

    @property
    def currency(self) -> str:
        return self.model.currency

    @property
    def state(self) -> float:
        return self.model.state

    @state.setter
    def state(self, val) -> None:
        if self.model.state != val:
            self.model.state = val

    @property
    def use_cent(self) -> bool:
        return self.model.use_cent

    @property
    def source(self) -> str:
        return self.model.source

    #not for peaqnext
    @property
    def average_month(self) -> float:
        return self.model.average_month

    #not for peaqnext
    @property
    def average_three_days(self) -> float:
        return self.model.average_three_days

    #not for peaqnext
    @property
    def average_weekly(self) -> float:
        return self.model.average_weekly

    #not for peaqnext
    @property
    def average_30(self) -> float:
        return self.model.average_30

    #not for peaqnext
    @property
    def average_data(self) -> dict[date, float]:
        return self.model.average_data
    
    @property
    def average_stdev_data(self) -> dict[date, float]:
        return self.model.average_stdev_data


    @abstractmethod
    async def async_set_dto(self, ret, initial: bool) -> None:
        pass

    @abstractmethod
    def setup(self):
        pass

    async def async_update_spotprice(self, initial: bool = False) -> None:
        if self.entity is not None:
            ret = self.state_machine.states.get(self.entity)
            if ret is not None:
                await self.async_set_dto(ret, initial)
            else:
                _LOGGER.debug(
                    f"Could not get spot-prices. Entity: {self.entity}. Retrying..."
                )

    async def async_update_set_prices(self, result: ISpotPriceDTO) -> bool:
        ret = False
        today = await self.model.fix_dst(result.today)
        if self.model.prices != today:
            self.model.prices = today if today else []
            ret = True
        if result.tomorrow_valid:
            self.model.tomorrow_valid = True
            tomorrow = await self.model.fix_dst(result.tomorrow)
            if self.model.prices_tomorrow != tomorrow:
                self.model.prices_tomorrow = tomorrow if tomorrow else []
                ret = True
        else:
            self.model.tomorrow_valid = False
            if self.model.prices_tomorrow:
                self.model.prices_tomorrow = []
                ret = True
        await self.async_update_average([3, 7, 30])
        self.model.currency = result.currency
        self.model.use_cent = result.price_in_cent
        self.state = result.state
        await self.async_update_average_day(result.average)
        await self.async_add_average_stdev_data(result.stdev)
        await self.async_update_average_month()
        await self.async_update_dynamic_max_price()
        return ret

    #not for peaqnext
    async def async_update_dynamic_max_price(self):
        if len(self.model.average_data) > 3:
            _dynamic_max_price = await self._dynamic_top_price.async_get_max(
                list(self.model.average_data.values())
            )
            if self.model.dynamic_top_price != _dynamic_max_price[0]:
                self.model.dynamic_top_price_type = _dynamic_max_price[1].value
                self.model.dynamic_top_price = _dynamic_max_price[0]
                await self.observer.async_broadcast(
                    ObserverTypes.DynamicMaxPriceChanged, _dynamic_max_price[0]
                )

    async def async_update_average_month(self) -> None:
        _new = self._get_average(datetime.now().day)
        if (
            len(self.model.average_data) >= int(datetime.now().day)
            and self.model.average_month != _new
        ):
            self.model.average_month = _new
            if self.model.average_month is not None:
                _LOGGER.debug(f"broadcasting average month: {_new}")
                await self.observer.async_broadcast(
                    ObserverTypes.MonthlyAveragePriceChanged, _new
                )

    async def async_update_average(self, length: list[int]) -> None:
        averages_dict = {
            3: "average_three_days",
            7: "average_weekly",
            30: "average_30",
        }
        for l in length:
            _new = self._get_average(l)
            if len(self.model.average_data) >= l and getattr(self.model, averages_dict[l]) != _new:
                setattr(self.model, averages_dict[l], _new)
                _LOGGER.debug(f"average {str(l)} updated to {_new}")
        await self.async_update_adjusted_average()

    async def async_update_adjusted_average(self) -> None:
        adj_avg: float|None = None
        if len(self.model.average_data) >= 7:
            adj_avg = max(self.model.average_weekly, self.model.average_three_days)
        elif len(self.model.average_data) >= 3:
            adj_avg = self.model.average_three_days
        if self.model.adjusted_average != adj_avg and adj_avg is not None:
            self.model.adjusted_average = adj_avg
            await self.observer.async_broadcast(
                ObserverTypes.AdjustedAveragePriceChanged, adj_avg
            )

    async def async_update_average_day(self, average) -> None:
        if average != self.model.daily_average:
            self.model.daily_average = average
            await self.async_add_average_data(average)
            await self.observer.async_broadcast(
                ObserverTypes.DailyAveragePriceChanged, average
            )

    async def async_import_average_data(self, incoming_prices: list|dict, incoming_stdev: list|dict|None = None):
        if len(incoming_prices):
            self.model.average_data = incoming_prices
        await self.async_cap_average_data_length(self.model.average_data)

        if incoming_stdev is not None and len(incoming_stdev):
            self.model.average_stdev_data = incoming_stdev
        await self.async_cap_average_data_length(self.model.average_stdev_data)
        await self.async_update_spotprice()

    async def async_add_average_data(self, new_val):
        if isinstance(new_val, float):
            rounded = round(new_val, 3)
            if datetime.now().date not in self.model.average_data.keys():
                self.model.average_data[date.today()] = rounded
            await self.async_cap_average_data_length(self.model.average_data)

    async def async_add_average_stdev_data(self, new_val):
        if isinstance(new_val, float):
            rounded = round(new_val, 3)
            if datetime.now().date not in self.model.average_stdev_data.keys():
                self.model.average_stdev_data[date.today()] = rounded
            await self.async_cap_average_data_length(self.model.average_stdev_data)

    async def async_cap_average_data_length(self, data: dict):
        while len(data) > AVERAGE_MAX_LEN:
            min_key = min(data.keys())
            del data[min_key]

    def _get_average(self, days: int) -> float:
        try:
            if len(self.model.average_data) > days:
                avg_values = list(self.model.average_data.values())
                ret = avg_values[-days:]
            elif len(self.model.average_data) == 0:
                return 0
            else:
                ret = list(self.model.average_data.values())
            return round(mean(ret), 2)
        except Exception as e:
            _LOGGER.debug(
                f"Could not calculate average. indata: {list(self.model.average_data.values())}, error: {e}"
            )
            return 0










import logging
from dataclasses import dataclass, field
from datetime import date, datetime, timedelta

_LOGGER = logging.getLogger(__name__)


@dataclass
class SpotPriceModel:
    source: str
    currency: str = ""
    prices: list = field(default_factory=lambda: [])
    prices_tomorrow: list = field(default_factory=lambda: [])
    state: float = 0
    entity: str = ""
    _average_data: dict = field(default_factory=lambda: {})
    _average_stdev_data: dict = field(default_factory=lambda: {})
    average_month: float = 0
    adjusted_average: float = 0
    average_weekly: float = 0
    average_three_days: float = 0
    average_30: float = 0
    tomorrow_valid: bool = False
    daily_average: float = 0
    use_cent: bool = False
    dynamic_top_price_type: str = ""
    dynamic_top_price: float|None = None


    @property
    def average_data(self) -> dict:
        return self._average_data
    
    @average_data.setter
    def average_data(self, value: dict|list) -> None:
        self._average_data = self.create_date_dict(value)

    @property
    def average_stdev_data(self) -> dict:
        return self._average_stdev_data
    
    @average_stdev_data.setter
    def average_stdev_data(self, value: dict|list) -> None:
        self._average_stdev_data = self.create_date_dict(value)

    def create_date_dict(self, numbers: dict|list) -> dict:
        if isinstance(numbers, dict):
            try:
                ret = {datetime.strptime(key, "%Y-%m-%d").date(): value for key, value in numbers.items()}
            except ValueError:
                _LOGGER.error("Could not convert date string to date object")
                ret = {}
        else:
            today = date.today()
            delta = timedelta(days=len(numbers) - 1)
            start_date = today - delta
            ret = {}
            for number in numbers:
                ret[start_date] = number
                start_date += timedelta(days=1)
        return ret

    async def fix_dst(self, val) -> list | None:
        if val is None:
            return None
        if len(val) < 1:
            return []
        ll = [h for h in val if h is not None]
        av = round(min(ll), 3)
        if len(val) == 23:
            val.insert(2, av)
        elif val[2] is None:
            val[2] = av
        return val



from .models.hour_price import HourPrice
from ...common.enums.cautionhourtype import CautionHourType
from .models.hour_type import HourType
from statistics import mean, stdev

LOCUTOFF = "lo_cutoff"
HICUTOFF = "hi_cutoff"
MAXHOURS = "max_hours"
MINHOURS = "min_hours"

def set_blank_permittance(hour_prices: list[HourPrice]) -> None:
    for hour in hour_prices:
        if hour.hour_type != HourType.AboveMax:
            hour.permittance = 1.0

def set_initial_permittance(
    hours: list[HourPrice],
    avg7: float | None = None,
    non_hours: list[int] = []
) -> None:
    avg = mean([h.price for h in hours if not h.passed])
    ceil = mean([avg, avg7]) if avg7 is not None else avg
    floor = min(avg, avg7) if avg7 is not None else 0
    for hour in hours:
        if not hour.passed:
            if hour.dt.hour in non_hours:
                hour.permittance = 0.0
            elif hour.price < floor:
                hour.permittance = 1.0
            elif hour.price > ceil:
                hour.permittance = 0.0
            else:
                hour.permittance = (hour.price < ceil) * (0.2 + 0.8 * (ceil - hour.price) / (ceil - floor))


def set_scooped_permittance(
    hour_prices: list[HourPrice], caution_hour_type: CautionHourType
) -> None:
    opt = _get_caution_options(caution_hour_type)
    for hp in hour_prices:
        if not hp.passed:
            if hp.permittance <= opt.get(LOCUTOFF, 0.4):
                hp.permittance = 0.0
            elif hp.permittance >= opt.get(HICUTOFF, 0.75):
                hp.permittance = 1.0
            else:
                hp.permittance = round(hp.permittance, 2)


def set_min_allowed_hours(
    hour_prices: list[HourPrice], caution_hour_type: CautionHourType
) -> None:
    opt = _get_caution_options(caution_hour_type)
    available_len = 0
    _min_hours = opt.get(MINHOURS, 4)
    available_len = len(
        [hp for hp in hour_prices if hp.permittance == 1 and not hp.passed]
    )
    if available_len < _min_hours:
        _t = [
            hp
            for hp in hour_prices
            if hp.hour_type != HourType.AboveMax
            and not hp.passed
            and hp.permittance < 1
        ]
        if len(_t):
            _t.sort(key=lambda x: (x.price,x.dt))
            for h in range(0,int(_min_hours - available_len)):
                _t[h].permittance = 1.0
    discard_peaks(hour_prices)

def _get_caution_options(caution_hour_type: CautionHourType, is_quarterly:bool = False) -> dict[str,float]:
    _quarters = 4 if is_quarterly else 1
    lo_cutoff = 0.5
    hi_cutoff = 0.8
    max_hours = 24 *_quarters
    min_hours = 4
    match caution_hour_type:
        case CautionHourType.SUAVE:
            lo_cutoff = 0.2
            hi_cutoff = 0.5
        case CautionHourType.INTERMEDIATE:
            lo_cutoff = 0.45
            hi_cutoff = 0.65
        case CautionHourType.AGGRESSIVE:
            lo_cutoff = 0.55
            max_hours = 20 * _quarters
        case CautionHourType.SCROOGE:
            lo_cutoff = 0.55
            max_hours = 8 * _quarters
            min_hours = 0
    return {
        LOCUTOFF: lo_cutoff,
        HICUTOFF: hi_cutoff,
        MAXHOURS: max_hours,
        MINHOURS: min_hours,
    }

def discard_peaks(hour_prices: list[HourPrice]) -> None:
    if _is_pointy(hour_prices):
        _set_peak_permittance_adjacent(hour_prices)
    else:
        _set_peak_permittance(hour_prices)

def _set_peak_permittance(hour_prices: list[HourPrice]) -> None:
    for i in range(1, len(hour_prices) - 1):
        if hour_prices[i].price > hour_prices[i-1].price and hour_prices[i].price > hour_prices[i+1].price:
            if hour_prices[i].hour_type != HourType.BelowMin:
                hour_prices[i].permittance = 0.0

def _set_peak_permittance_adjacent(hour_prices: list[HourPrice]) -> None:
    prices = [hour.price for hour in hour_prices]
    std_dev = stdev(prices)
    for i in range(1, len(hour_prices) - 1):
        if hour_prices[i].price > hour_prices[i-1].price and hour_prices[i].price > hour_prices[i+1].price:
            if hour_prices[i].price - hour_prices[i-1].price > std_dev and hour_prices[i].price - hour_prices[i+1].price > std_dev:
                if hour_prices[i].hour_type != HourType.BelowMin:
                    hour_prices[i].permittance = 0.0 
                if hour_prices[i-1].hour_type != HourType.BelowMin:
                    hour_prices[i-1].permittance = 0.0
                if hour_prices[i+1].hour_type != HourType.BelowMin:
                    hour_prices[i+1].permittance = 0.0

def _is_pointy(hour_prices:list[HourPrice]) -> bool:
    _prices = [hp.price for hp in hour_prices]
    return mean(_prices) < max(_prices) - min(_prices)


from datetime import date, time, timedelta, datetime
import logging
from .models.stop_string import AllowanceObj, set_allowance_obj
from .models.datetime_model import DateTimeModel
from .models.hour_price import HourPrice
from .models.hourselection_service_model import HourSelectionServiceModel
from ...models.hourselection.hourselection_options import HourSelectionOptions
from .hourselection_calculations import normalize_prices, do_recalculate_prices, get_average_kwh_price
from .permittance import (
    set_initial_permittance,
    set_scooped_permittance,
    set_blank_permittance,
    set_min_allowed_hours,
)
from statistics import stdev, mean
from .max_min_charge import MaxMinCharge

_LOGGER = logging.getLogger(__name__)

class HourSelectionService:
    def __init__(self, options: HourSelectionOptions = HourSelectionOptions()):
        self.options = options
        self.dtmodel = DateTimeModel()
        self.model = HourSelectionServiceModel()
        self.max_min = MaxMinCharge(service=self, min_price=self.options.min_price)

    @property
    def all_hours(self) -> list[HourPrice]:
        self.update()
        return self.model.hours_prices

    @property
    def future_hours(self) -> list[HourPrice]:
        self.update()
        return self.model.get_future_hours(self.dtmodel)

    @property
    def display_future_hours(self) -> list[HourPrice]:
        if self.max_min.active and not self.max_min.overflow:
            return self.max_min.future_hours(self.dtmodel)
        return self.future_hours

    @property
    def allowance(self) -> AllowanceObj:
        self.update()
        return set_allowance_obj(self.dtmodel, self.display_future_hours)

    @property
    def average_kwh_price(self) -> float:
        return round(get_average_kwh_price(self.future_hours), 2)

    @property
    def offset_dict(self) -> dict:
        return self.model.get_offset_dict(self.dtmodel.dt)

    def update(self):  
        self.model.hours_prices = [hp for hp in self.model.hours_prices if hp.dt.date() >= self.dtmodel.hdate]        
        for hp in self.model.hours_prices:            
            hp.set_passed(self.dtmodel)        
        if len(self.model.get_future_hours(self.dtmodel)) >= 24:
            set_scooped_permittance(
                self.model.get_future_hours(self.dtmodel), self.options.cautionhour_type_enum
            )        

    async def async_update(self):
        self.update()

    def clean_prices(self, prices, prices_tomorrow) -> dict:
        hours_today = self.dtmodel.get_eligable_hours(self.dtmodel.hdate)
        price_dict: dict = {}
        for idx, h in enumerate(hours_today):
            try:
                price_dict[h] = prices[idx]
            except Exception as e:
                print(f"Exception in adding prices today! {e}")
        if len(prices_tomorrow):
            hours_tomorrow = self.dtmodel.get_eligable_hours(self.dtmodel.hdate_tomorrow)
            for idx, h in enumerate(hours_tomorrow):
                try:
                    price_dict[h] = prices_tomorrow[idx]
                except Exception as e:
                    print(f"Exception in adding prices tomorrow! {e}")
        return price_dict

    async def async_update_prices(self, prices: list[float], prices_tomorrow: list[float] = []):
        price_dict = self.clean_prices(prices, prices_tomorrow)
        self.model.prices_today = prices  # clean first
        self.model.prices_tomorrow = prices_tomorrow  # clean first
        if do_recalculate_prices(price_dict=price_dict, hours_prices=self.model.hours_prices, hdate=self.dtmodel.hdate):
            print(f"do recalculate prices {len(prices)} {len(prices_tomorrow)}")
            _LOGGER.debug(f"do recalculate prices {len(prices)} {len(prices_tomorrow)}")
            self._create_prices(prices, prices_tomorrow)
        else:
            print(f"NOT recalculate prices {len(prices)} {len(prices_tomorrow)}")
            _LOGGER.debug(f"NOT recalculate prices {len(prices)} {len(prices_tomorrow)}")
            await self.async_update()
        self.max_min.get_hours()

    async def async_update_adjusted_average(self, adjusted_average: float) -> None:
        self.model.adjusted_average = adjusted_average
        self.update()
        if len(self.model.hours_prices) > 0:
            self._set_permittance()
        else:
            await self.async_update()

    def _create_prices(
        self, prices: list[float], prices_tomorrow: list[float] = []
    ) -> None:
        # todo: fix to allow 23, 24,25, 92, 96, 100 for dst-dates.
        self.model.hours_prices = []
        match len(prices):
            case 23 | 24 | 25:
                self._create_hour_prices(prices, prices_tomorrow, is_quarterly=False)
            case 92 | 96 | 100:
                self._create_hour_prices(prices, prices_tomorrow, is_quarterly=True)
            case _:
                raise ValueError(
                    f"Length of pricelist must be either 23,24,25,92,96 or 100. Your length is {len(prices)}"
                )

    def _create_hour_prices(
        self,
        prices: list[float],
        prices_tomorrow: list[float] = [],
        is_quarterly: bool = False,
    ) -> None:
        # todo: handle here first if prices or prices_tomorrow are 92 or 100 in length (dst shift)
        self.model.use_quarters = is_quarterly
        self.model.set_hourprice_lists(prices, prices_tomorrow, self.options, self.dtmodel.hdate, self.dtmodel.hdate_tomorrow, self.dtmodel.is_passed, self.dtmodel.get_eligable_hours)
        self._set_permittance()

    def _set_permittance(self) -> None:
        prices = normalize_prices(
            [
                hp.price
                for hp in self.model.hours_prices
                if hp.dt.date() >= self.dtmodel.hdate
            ]
        )
        self.model.set_offset_dict(prices, self.dtmodel.dt.date(), self.options.min_price)
        if not self.is_flat([h.price for h in self.model.hours_prices if not h.passed]):
            set_initial_permittance(
                self.model.hours_prices,
                self.model.adjusted_average,
                self.options.non_hours
            )
        else:
            set_blank_permittance(self.model.hours_prices)
        set_scooped_permittance(
            self.model.hours_prices,
            self.options.cautionhour_type_enum,
        )
        set_min_allowed_hours(
            self.model.hours_prices,
            self.options.cautionhour_type_enum,
        )

    @staticmethod
    def is_flat(prices: list[float]) -> bool:
        """if the fsd is below a score of 8(%) it is considered flat"""
        return stdev(prices)*100/mean(prices) < 8


from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ...services.locale.querytypes.queryservice import QueryService
from dataclasses import dataclass, field
from datetime import datetime
from ...models.locale.enums.querytype import QueryType
from ...models.locale.enums.time_periods import TimePeriods
from ..locale.time_pattern import TimePattern
from ...models.locale.price.locale_price import LocalePrice
from .locale_query import LocaleQuery, ILocaleQuery
from abc import abstractmethod


class Locale_Type:
    _observed_peak = None
    _charged_peak = None
    _query_model: ILocaleQuery = field(default_factory=ILocaleQuery)
    _query_service = None
    _price: LocalePrice = LocalePrice()
    _free_charge_pattern = None
    _peak_cycle: TimePeriods = TimePeriods.Hourly

    @property
    @abstractmethod
    def observed_peak(self) -> QueryType:
        return self._observed_peak

    @observed_peak.setter
    @abstractmethod
    def observed_peak(self, value: QueryType):
        self._observed_peak = value

    @property
    @abstractmethod
    def charged_peak(self) -> QueryType:
        return self._charged_peak

    @charged_peak.setter
    @abstractmethod
    def charged_peak(self, value: QueryType):
        self._charged_peak = value

    @property
    @abstractmethod
    def query_model(self) -> ILocaleQuery:
        return self._query_model

    
    @query_model.setter
    @abstractmethod
    def query_model(self, value: ILocaleQuery):
        self._query_model = value

    @property
    @abstractmethod
    def query_service(self) -> QueryService:
        return self._query_service

    @query_service.setter
    @abstractmethod
    def query_service(self, value: QueryService):
        self._query_service = value

    @property
    @abstractmethod
    def price(self) -> LocalePrice:
        return self._price

    @price.setter
    @abstractmethod
    def price(self, value: LocalePrice):
        self._price = value

    @property
    @abstractmethod
    def free_charge_pattern(self) -> TimePattern:
        return self._free_charge_pattern

    @free_charge_pattern.setter
    @abstractmethod
    def free_charge_pattern(self, value: TimePattern):
        self._free_charge_pattern = value

    @property
    def peak_cycle(self) -> TimePeriods:
        return self._peak_cycle

    @peak_cycle.setter
    def peak_cycle(self, value: TimePeriods):
        self._peak_cycle = value

    async def async_free_charge(self, mockdt: datetime | None = None) -> bool:
        if self.free_charge_pattern is None:
            return False
        now = mockdt or datetime.now()
        return await self.free_charge_pattern.async_valid(now)

    async def async_is_quarterly(self) -> bool:
        return self.peak_cycle == TimePeriods.QuarterHourly

    def is_quarterly(self) -> bool:
        return self.peak_cycle == TimePeriods.QuarterHourly

    async def async_set_query_service(self):
        if self.query_service is not None:
            await self.query_model.async_set_query_service(self.query_service)



import time

class WaitTimer:
    def __init__(self, timeout: int, init_now: bool = True):
        self._timeout = timeout
        self._base_timeout = timeout
        self._last_update = time.time() if init_now else 0

    @property
    def timeout(self) -> int:
        return self._timeout

    @property
    def value(self) -> float:
        return self._last_update

    def update(self, override=None) -> None:
        if override is not None:
            self._timeout = override
        self._last_update = time.time()

    def reset(self) -> None:
        self._last_update = 0
        self._timeout = self._base_timeout

    def is_timeout(self) -> bool:
        if self._timeout == 0:
            return False
        if time.time() - self._last_update > self._timeout:
            self._last_update = time.time()
            self._timeout = self._base_timeout
            return True
        return False
    

boost_wait = WaitTimer(timeout=300, init_now=False)
print(boost_wait.is_timeout())
boost_wait.update()
print(boost_wait.is_timeout())



import logging
from typing import Tuple
from ...common.enums.cautionhourtype import CautionHourType
from ...models.hourselection.hourselection_model import HourSelectionModel
from ...models.hourselection.hourselection_options import HourSelectionOptions
from ..hoursselection_service_new.hourselection_service import HourSelectionService

from datetime import datetime

_LOGGER = logging.getLogger(__name__)


class Hoursselection:
    def __init__(
        self,
        absolute_top_price: float = 0,
        min_price: float = 0,
        cautionhour_type: str | CautionHourType = CautionHourType.SUAVE.value,
        non_hours: list[int] = []
    ):
        self.cautionhour_type_enum = (
            CautionHourType(cautionhour_type.lower())
            if isinstance(cautionhour_type, str)
            else cautionhour_type
        )
        self.model = HourSelectionModel(
            options=HourSelectionOptions(
                cautionhour_type_enum=self.cautionhour_type_enum,
                min_price=min_price,
                top_price=absolute_top_price,
                non_hours=non_hours
            )
        )
        self.model.validate()
        self.service = HourSelectionService(options=self.model.options)

    @property
    def offsets(self) -> dict:
        ret = {}
        ret["today"] = self.service.offset_dict["today"]
        ret["tomorrow"] = self.service.offset_dict["tomorrow"]
        return ret

    @property
    def caution_hours(self) -> list[datetime]:
        return sorted(list(self.dynamic_caution_hours.keys()))

    @property
    def non_hours(self) -> list[datetime]:
        self.service.update()
        return [hp.dt for hp in self.future_hours if hp.permittance == 0.0]

    @property
    def dynamic_caution_hours(self) -> dict[datetime, float]:
        self.service.update()
        ret = {
            hp.dt: hp.permittance
            for hp in self.future_hours
            if 0.0 < hp.permittance < 1.0
        }
        keys = list(ret.keys())
        keys.sort()
        return {k: ret[k] for k in keys}

    @property
    def future_hours(self) -> list:
        return self.service.display_future_hours

    @property
    def prices(self) -> list:
        return self.service.model.prices_today

    @property
    def prices_tomorrow(self) -> list:
        return self.service.model.prices_tomorrow

    @property
    def adjusted_average(self):
        return self.model.adjusted_average

    @adjusted_average.setter
    def adjusted_average(self, val):
        if val != self.model.adjusted_average:
            self.model.adjusted_average = val

    async def async_update_adjusted_average(self, val):
        await self.service.async_update_adjusted_average(val)
        
    async def async_update_top_price(self, dyn_top_price=None) -> None:
        await self.model.options.async_set_absolute_top_price(dyn_top_price)
        self.service.model.update_hour_types(self.model.options)

    async def async_update_prices(self, prices: list, prices_tomorrow: list = []):
        await self.service.async_update_prices(prices, prices_tomorrow)

    async def async_get_average_kwh_price(self) -> Tuple[float | None, float | None]:
        ret_static = self.service.average_kwh_price
        ret_dynamic = self.service.max_min.average_price
        if ret_dynamic is not None:
            if ret_dynamic > ret_static:
                ret_dynamic = ret_static
        return ret_static, ret_dynamic

    async def async_get_total_charge(
        self, currentpeak: float
    ) -> Tuple[float, float | None]:
        ret_dynamic = self.service.max_min.total_charge
        self.model.current_peak = currentpeak
        _charge = (
            self.service.max_min.model.expected_hourly_charge
            if self.service.max_min.active
            else currentpeak
        )
        ret_static = round(
            sum([hp.permittance * _charge for hp in self.service.future_hours]),
            1,
        )
        if ret_dynamic is not None:
            if ret_dynamic > ret_static:
                ret_dynamic = ret_static
        return ret_static, ret_dynamic



from statistics import mean, stdev
import logging
from .models.hour_price import HourPrice
from datetime import date

_LOGGER = logging.getLogger(__name__)


def normalize_prices(prices: list) -> list:
    min_price = min(prices)
    if min_price > 0:
        return prices
    c = 0
    if min_price <= 0:
        c = abs(min_price) + 0.01
    ret = []
    for p in prices:
        pp = p + c
        divider = min_price if min_price > 0 else c
        ret.append(round(pp - divider, 3))
    return ret

def do_recalculate_prices(price_dict: dict, hours_prices: list[HourPrice], hdate: date) -> bool:
    if len(hours_prices) < 1:
        return True
    for d in price_dict.items():
        if d[0] not in [hp.dt for hp in hours_prices]:
            # print(f"{d[0]} is not in hours_prices")
            # print(f"hours_prices is {hours_prices}")
            return True
    return False
    
# def do_recalculate_prices(price_dict: dict, hours_prices: list[HourPrice], hdate: date) -> bool:
#         if any([hp.dt.date() == hdate for hp in hours_prices]):
#         if [
#             hp.price
#             for hp in hours_prices
#             if hp.dt.date() == hdate
#         ] == prices and len(prices_tomorrow) < 1:
#             print("prices are the same as the previous day, so we will not recalculate")
#             """We should not recalculate prices if they are the same as the previous day"""
#             return False
#         return True


def get_average_kwh_price(future_hours: list[HourPrice]) -> float:
        try:
            return mean(
                [
                    hp.price
                    for hp in future_hours
                    if hp.permittance > 0
                ]
            )
        except Exception:
            return 0.0


from ....services.locale.querytypes.queryservice import QueryService
from ....models.locale.enums.querytype import QueryType
from ..time_pattern import TimePattern
from ....models.locale.price.locale_price import LocalePrice
from ....models.locale.price.models.seasoned_price import SeasonedPrice
from ....models.locale.enums.price_type import PriceType
from ....models.locale.enums.calendar_periods import CalendarPeriods
from ..querytypes.const import (
    QUERYTYPE_BASICMAX_MON_FRI_07_17_DEC_MAR_ELSE_REGULAR,
    QUERYTYPE_SOLLENTUNA,
    QUERYTYPE_MAX_NOV_MAR_MON_FRI_06_22,
    QUERYTYPE_AVERAGEOFTHREEHOURS_MON_FRI_07_19,
)

from ..querytypes.querysets import QUERYSETS
from ..querytypes.querytypes import QUERYTYPES
from dataclasses import dataclass
from ..locale_model import Locale_Type


@dataclass
class SE_Sollentuna(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.AverageOfThreeHours
        self.charged_peak = QueryType.AverageOfThreeHours
        self.query_model = QUERYTYPES[QueryType.AverageOfThreeHours]
        self.query_service = QueryService(QUERYSETS[QUERYTYPE_SOLLENTUNA])
        self.free_charge_pattern = TimePattern(
            [
                {
                    CalendarPeriods.Month: [*range(1, 13)],
                    CalendarPeriods.Weekday: [0, 1, 2, 3, 4],
                    CalendarPeriods.Hour: [19, 20, 21, 22, 23, 0, 1, 2, 3, 4, 5, 6],
                },
                {
                    CalendarPeriods.Month: [*range(1, 13)],
                    CalendarPeriods.Weekday: [5, 6],
                    CalendarPeriods.Hour: [*range(0, 24)],
                },
            ]
        )

    # Rörlig avgift sommar april – oktober 61,46 kr/kW
    # Rörlig avgift vinter november – mars 122,92 kr/kW
    # https://www.seom.se/el/elnat/2022-ars-priser-och-villkor/


@dataclass
class SE_Skovde(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.Max
        self.charged_peak = QueryType.Max
        self.query_model = QUERYTYPES[QueryType.Max]
        self.query_service = QueryService(
            QUERYSETS[QUERYTYPE_MAX_NOV_MAR_MON_FRI_06_22]
        )
        self.free_charge_pattern = TimePattern(
            [
                {
                    CalendarPeriods.Month: [11, 12, 1, 2, 3],
                    CalendarPeriods.Weekday: [5, 6],
                    CalendarPeriods.Hour: [22, 23, 0, 1, 2, 3, 4, 5],
                },
                {
                    CalendarPeriods.Month: [4, 5, 6, 7, 8, 9, 10],
                    CalendarPeriods.Weekday: [*range(0, 7)],
                    CalendarPeriods.Hour: [*range(0, 24)],
                },
            ]
        )

    # November-Mars, vardagar (mån-fre) 06-22
    # single peak i denna period månadsvis.


@dataclass
class SE_SHE_AB(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.AverageOfThreeHours
        self.charged_peak = QueryType.AverageOfThreeHours
        self.query_model = QUERYTYPES[QueryType.AverageOfThreeHours]
        self.query_service = QueryService(
            QUERYSETS[QUERYTYPE_AVERAGEOFTHREEHOURS_MON_FRI_07_19]
        )
        self.free_charge_pattern = TimePattern(
            [
                {
                    CalendarPeriods.Month: [*range(1, 13)],
                    CalendarPeriods.Weekday: [5, 6],
                    CalendarPeriods.Hour: [19, 20, 21, 22, 23, 0, 1, 2, 3, 4, 5, 6],
                }
            ]
        )

    """
    Elnätskunder med effekttaxa får vinterpris på överföringsavgift från och med 1 november – 31 mars. 
    Prishöjningen på effekttaxan skedde 1 april men blir mer kännbart när vinterpriset nu träder i kraft. 
    Kunder som bor i villa och har effekttaxa kan påverka sin kostnad genom att försöka skjuta på sådan förbrukning 
    som är möjlig från dagtid till kvällstid (19.00-07:00) eller till helger och röda dagar då det är helt kostnadsfritt att använda elnätet.
    Överföringsavgiften beräknas på medelmånadsvärdet av de tre högsta effektvärden dagtid vardagar mellan 07.00-19.00.
    
    Nov – Mars vardagar kl 7-19 135,00 kr/kW inkl moms
    April – Okt vardagar kl 7-19 56,00 kr/kW inkl moms
    """


@dataclass
class SE_Partille(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.Max
        self.charged_peak = QueryType.Max
        self.query_model = QUERYTYPES[QueryType.Max]
        self.price = LocalePrice(
            price_type=PriceType.Static, value=29.5, currency="SEK"
        )


@dataclass
class SE_FalbygdensEnergi(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.Max
        self.charged_peak = QueryType.Max
        self.query_model = QUERYTYPES[QueryType.Max]


@dataclass
class SE_Nacka_normal(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.AverageOfThreeHours
        self.charged_peak = QueryType.AverageOfThreeHours
        self.query_model = QUERYTYPES[QueryType.AverageOfThreeHours]

    # https://www.nackaenergi.se/images/downloads/natavgifter/FAQ_NYA_TARIFFER.pdf


@dataclass
class SE_NACKA_timediff(Locale_Type):
    pass

    # this class is for nacka time differentiated peaks.
    # https://www.nackaenergi.se/images/downloads/natavgifter/FAQ_NYA_TARIFFER.pdf


@dataclass
class SE_Malung_Salen(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.AverageOfFiveDays
        self.charged_peak = QueryType.AverageOfFiveDays
        self.query_model = QUERYTYPES[QueryType.AverageOfFiveDays]
        self.free_charge_pattern = TimePattern(
            [
                {
                    CalendarPeriods.Month: [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
                    CalendarPeriods.Weekday: [0, 1, 2, 3, 4, 5, 6],
                    CalendarPeriods.Hour: [19, 20, 21, 22, 23, 0, 1, 2, 3, 4, 5, 6],
                }
            ]
        )

    # Rörlig avgift sommar april – oktober 35 kr/kW
    # Rörlig avgift vinter november – mars 118,75 kr/kW


@dataclass
class SE_Linde_Energi(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.Max
        self.charged_peak = QueryType.Max
        self.query_model = QUERYTYPES[QueryType.Max]

    # docs: https://www.lindeenergi.se/elnat/elnatspriser/effekttariffer.4.1491a0b016e44ba6ccfe91b4.html

    """
    Din effektavgift baseras på din högsta effekttopp per månad, alltså den timme per månad då du använder mest el sammantaget, oavsett tid på dygnet. 
    Under perioden november till mars tillkommer en högbelastningslastavgift, så kallad höglasteffekt, då du även debiteras för ditt högst uppmätta timvärde vardagar kl 07:00-19:00.
    Under vintermånaderna kan du alltså debiteras för två olika timvärden, beroende på när på dygnet din effekttopp uppmäts.
    """


@dataclass
class SE_Kristinehamn(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.Max
        self.charged_peak = QueryType.Max
        self.query_model = QUERYTYPES[QueryType.Max]
        self.query_service = QueryService(
            QUERYSETS[QUERYTYPE_BASICMAX_MON_FRI_07_17_DEC_MAR_ELSE_REGULAR]
        )

    """
    https://kristinehamnsenergi.se/elnat/elnatsavgiften/effektavgift-villa-med-bergvarmepump/
    vardagar november-mars, kl 07.00-17.00 > highload instead of normal load. other times, normal load
    """


@dataclass
class SE_Karlstad(Locale_Type):
    # docs: https://karlstadsnat.se/elnat/kund/priser-och-tariffer/effekttariff/
    """
    Note, high load extra is added on weekdays from 6-18 during november - march.
    This does not affect the peak, but should in future updates be cause for forced non-/or caution-hours to lessen the cost for the consumer.
    """

    def __post_init__(self):
        self.observed_peak = QueryType.Max
        self.charged_peak = QueryType.Max
        self.query_model = QUERYTYPES[QueryType.Max]


@dataclass
class SE_Gothenburg(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.AverageOfThreeDays
        self.charged_peak = QueryType.AverageOfThreeDays
        self.query_model = QUERYTYPES[QueryType.AverageOfThreeDays]
        self.price = LocalePrice(
            price_type=PriceType.Static, value=43.75, currency="SEK"
        )


@dataclass
class SE_Ellevio(Locale_Type):
    # docs: https://www.ellevio.se/privat/om-din-el/forsta-din-elkostnad/ny-prismodell-for-elnatet/

    def __post_init__(self):
        self.observed_peak = QueryType.AverageOfThreeDays
        self.charged_peak = QueryType.AverageOfThreeDays
        self.query_model = QUERYTYPES[QueryType.AverageOfThreeDays]
        self.price = LocalePrice(price_type=PriceType.Static, value=65, currency="SEK")


@dataclass
class SE_JBF(Locale_Type):
    # docs: http://www.jbf.nu/sv/natavgift
    # Högbelastningstid är vardagar kl. 06.00 – 22.00 under månaderna november – mars.
    def __post_init__(self):
        self.observed_peak = QueryType.AverageOfThreeHours
        self.charged_peak = QueryType.AverageOfThreeHours
        self.query_model = QUERYTYPES[QueryType.AverageOfThreeHours]
        self.query_service = QueryService(
            QUERYSETS[QUERYTYPE_MAX_NOV_MAR_MON_FRI_06_22]
        )
        self.price = LocalePrice(
            price_type=PriceType.Static, value=96.6, currency="SEK"
        )
        self.free_charge_pattern = TimePattern(
            [
                {
                    CalendarPeriods.Month: [4, 5, 6, 7, 8, 9, 10],
                    CalendarPeriods.Weekday: [*range(0, 7)],
                    CalendarPeriods.Hour: [*range(0, 24)],
                },
                {
                    CalendarPeriods.Month: [11, 12, 1, 2, 3],
                    CalendarPeriods.Weekday: [5, 6],
                    CalendarPeriods.Hour: [*range(0, 24)],
                },
                {
                    CalendarPeriods.Month: [11, 12, 1, 2, 3],
                    CalendarPeriods.Weekday: [0, 1, 2, 3, 4],
                    CalendarPeriods.Hour: [22, 23, 0, 1, 2, 3, 4, 5],
                },
            ]
        )


@dataclass
class SE_Bjerke_Energi(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.Max
        self.charged_peak = QueryType.Max
        self.query_model = QUERYTYPES[QueryType.Max]
        self.price = LocalePrice(
            price_type=PriceType.Seasoned,
            currency="SEK",
            _values=[
                SeasonedPrice(
                    validity=TimePattern(
                        [
                            {
                                CalendarPeriods.Month: [11, 12, 1, 2, 3],
                                CalendarPeriods.Weekday: [0, 1, 2, 3, 4, 5, 6],
                                CalendarPeriods.Hour: [
                                    6,
                                    7,
                                    8,
                                    9,
                                    10,
                                    11,
                                    12,
                                    13,
                                    14,
                                    15,
                                    16,
                                    17,
                                    18,
                                    19,
                                    20,
                                    21,
                                ],
                            }
                        ]
                    ),
                    value=106.25,
                ),
                SeasonedPrice(
                    validity=TimePattern(
                        [
                            {
                                CalendarPeriods.Month: [4, 5, 6, 7, 8, 9, 10],
                                CalendarPeriods.Weekday: [0, 1, 2, 3, 4, 5, 6],
                                CalendarPeriods.Hour: [
                                    6,
                                    7,
                                    8,
                                    9,
                                    10,
                                    11,
                                    12,
                                    13,
                                    14,
                                    15,
                                    16,
                                    17,
                                    18,
                                    19,
                                    20,
                                    21,
                                ],
                            }
                        ]
                    ),
                    value=50,
                ),
            ],
        )

        self.free_charge_pattern = TimePattern(
            [
                {
                    CalendarPeriods.Month: [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
                    CalendarPeriods.Weekday: [0, 1, 2, 3, 4, 5, 6],
                    CalendarPeriods.Hour: [22, 23, 0, 1, 2, 3, 4, 5],
                }
            ]
        )

    # docs: https://www.bjerke-energi.se/elnat/tariffer/effekttariff-fr-o-m-2022-02-01/
    # dag kl. 06-22 nov-mars                 106,25 kr/kW/mån
    # dag kl. 06-22 april-okt                50 kr/kW/mån
    # natt kl. 22-06 alla dagar hela året    0 kr/kW/mån
    """
    Note, high load extra is added from 06-22 during november - march. 
    This does not affect the peak, but should in future 
    """


@dataclass
class SE_Telge_Energi(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.Max
        self.charged_peak = QueryType.Max
        self.query_model = QUERYTYPES[QueryType.Max]
        self.free_charge_pattern = TimePattern(
            [
                {
                    CalendarPeriods.Month: [4, 5, 6, 7, 8, 9, 10],
                    CalendarPeriods.Weekday: [0, 1, 2, 3, 4, 5, 6],
                    CalendarPeriods.Hour: [
                        21,
                        22,
                        23,
                        0,
                        1,
                        2,
                        3,
                        4,
                        5,
                        6,
                        7,
                        8,
                        9,
                        10,
                        11,
                        12,
                        13,
                        14,
                        15,
                        16,
                        17,
                        18,
                        19,
                        20,
                    ],
                },
                {
                    CalendarPeriods.Month: [11, 12, 1, 2, 3],
                    CalendarPeriods.Weekday: [5, 6],
                    CalendarPeriods.Hour: [21, 22, 23, 0, 1, 2, 3, 4, 5, 6],
                },
            ]
        )

    # docs: https://nya.telge.se/elnat/ny-prismodell-elnat/om-hog--och-laglasttid/

    """
    Det högre priset gäller bara på helgfria vardagar 07.00-20.00 under november till mars.
    """


@dataclass
class SE_Malarenergi(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.AverageOfThreeHours
        self.charged_peak = QueryType.AverageOfThreeHours
        self.query_model = QUERYTYPES[QueryType.AverageOfThreeHours]

    # docs: https://www.malarenergi.se/el/elnat/effekttariffer/

    """
    Det högre priset gäller bara på helgfria vardagar 07.00-20.00 under november till mars.
    """


@dataclass
class SE_TekniskaVerken_Link(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.Max
        self.charged_peak = QueryType.Max
        self.query_model = QUERYTYPES[QueryType.Max]

    # docs: unknown


@dataclass
class SE_Eskilstuna(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.Max
        self.charged_peak = QueryType.Max
        self.query_model = QUERYTYPES[QueryType.Max]

    # docs: unknown
